package org.netkernel.module.standard;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.atomic.AtomicBoolean;

import org.netkernel.container.IKernel;
import org.netkernel.container.ILogger;
import org.netkernel.layer0.meta.IEndpointStateMeta;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.meta.ISpaceMeta;
import org.netkernel.layer0.meta.impl.SpaceElementsImpl;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.nkf.impl.*;
import org.netkernel.layer0.util.GoldenThreadExpiryFunction;
import org.netkernel.layer0.util.TupleList;
import org.netkernel.layer0.util.Utils;
import org.netkernel.layer0.util.XMLUtils;
import org.netkernel.module.standard.endpoint.IStandardEndpoint;
import org.netkernel.request.IRequest;
import org.netkernel.request.IRequestResponseFields;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.IResponseMeta;
import org.netkernel.request.impl.ExpiryFactory;
import org.netkernel.request.impl.RequestResponseFieldsImpl;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.scheduler.IInternalResponseMeta;
import org.netkernel.urii.ClassLoaderWithExports;
import org.netkernel.urii.IEndpoint;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.INetKernelException;
import org.w3c.dom.Element;

public class StandardSpace extends NKFCombinedEndpointSpaceImpl
{	
	private final StandardSpace mParentSpace;
	private final StandardModule mOwner;
	private final Element mConfig;
	private final IRequestScopeLevel mStaticScope;
	private final StandardSpace[] mStaticScopeArray;
	private INKFResponseReadOnly mLastMetaResponse;
	private AtomicBoolean mBindingLock=new AtomicBoolean();
	private boolean mIsBound;
	private String mName;
	protected IStandardEndpoint[] mMappings;
	private IRequestResponseFields mState=RequestResponseFieldsImpl.EMPTY;
	private int mHash;
	private GoldenThreadExpiryFunction mInvalidationExpiry;
	
	public StandardSpace(StandardSpace aParentSpace, StandardModule aOwner, Element aConfig)
	{
		declareThreadSafe();
		mParentSpace=aParentSpace;
		mOwner=aOwner;
		mConfig=aConfig;
		mHash=System.identityHashCode(this);
		mName="";
		
		// build static scope of space
		List<StandardSpace> spaces = new ArrayList<StandardSpace>();
		StandardSpace space = this;
		while (space!=null)
		{	spaces.add(space);
			space=space.getParent();
		}
		IRequestScopeLevel scope = RequestScopeLevelImpl.createRootScopeLevel(spaces.get(spaces.size()-1));
		mStaticScopeArray=new StandardSpace[spaces.size()];
		int i=spaces.size()-1;
		mStaticScopeArray[i]=spaces.get(i);
		for (i=spaces.size()-2; i>=0; i--)
		{	space = spaces.get(i);
			scope = RequestScopeLevelImpl.createNonDurableScopeLevel(space, scope);
			mStaticScopeArray[i]=space;
		}
		mStaticScope=scope;		
		
		invalidateMeta(false,false);
	}
	
	public StandardSpace getProtoSpace()
	{	return (mState==null)?null:(StandardSpace)mState.getValue(ISpaceMeta.STATE_PROTOSPACE);
	}
	
	public void releaseReferences()
	{	mLastMetaResponse=null;
		mState=null;
	}
	
	public int hashCode()
	{	return mHash;
	}
	
	public void setName(String aName)
	{	mName=aName;
	}
	
	public String getName()
	{	return mName;
	}
	
	public void setMappings(IStandardEndpoint[] aMappings)
	{	mMappings=aMappings;
	}
	
	public Element getConfigElement()
	{	return mConfig;
	}
		
	public StandardSpace getParent()
	{	return mParentSpace;
	}
	
	public IKernel getKernel()
	{	return super.getKernel();
	}
	
	public String getOwnerName()
	{	return mOwner.getMeta().getIdentifier();
	}
	
	public StandardModule getOwningModule()
	{	return mOwner;
	}
	
	
	public int getIndexOfMapping(IStandardEndpoint aMapping)
	{	int result=0;
		if (mMappings!=null)
		{	for (int i=0; i<mMappings.length; i++)
			{	if (aMapping==mMappings[i])
				{	result=i;
					break;
				}
			}
		}
		return result;
	}	

	public ClassLoaderWithExports getClassLoader()
	{	return mOwner.getClassLoader();
	}

	public void onCommissionSpace(IKernel aKernel) throws Exception
	{	super.onCommissionSpace(aKernel);
	
		IKernel kernel=aKernel;
		int errors=0;
		for (int i=0; i<mMappings.length; i++)
		{	IStandardEndpoint mapping = mMappings[i];
			try
			{	mapping.onCommissionEndpoint(kernel, this);
			}
			catch (Throwable e)
			{	mMappings[i]=null;
				errors++;
				String toString;
				ILogger logger=getKernel().getLogger();
				try
				{	toString = mapping.toString();
					if (toString==null)
					{	toString=mapping.getClass().getName();
					}
				} catch (Exception e2)
				{	toString=mapping.getClass().getName();
					logger.logRaw(ILogger.WARNING, this, Utils.throwableToString(e2));
				}
				String msg=logger.format("MSG_STD_ENDPOINT_COMMISSION_FAILED",mName,toString,Utils.throwableToString(e));
				logger.logRaw(ILogger.WARNING,this,msg);
				setInitialisationProblem(msg);
			}
		}
		
		if (errors>0)
		{	IStandardEndpoint[] newMappings = new IStandardEndpoint[mMappings.length-errors];
			int j=0;
			for (int i=0; i<mMappings.length; i++)
			{	IStandardEndpoint mapping = mMappings[i];
				if (mapping!=null)
				{	newMappings[j++]=mapping;
				}
			}
			mMappings=newMappings;
		}	
		
	}

	
	void setInitialisationProblem(String aProblem)
	{	setState(IEndpointStateMeta.KEY_PROBLEM, aProblem);
	}
	
	void clearInitialisationProblems()
	{	if (mState!=RequestResponseFieldsImpl.EMPTY)
		{	RequestResponseFieldsImpl state=(RequestResponseFieldsImpl)mState;
			while(state.getValue(IEndpointStateMeta.KEY_PROBLEM)!=null)
			{	state.remove(IEndpointStateMeta.KEY_PROBLEM);
			}
			if (state.size()==0)
			{	mState=RequestResponseFieldsImpl.EMPTY;
			}
		}
	}
	
	protected void setState(String aKey, Object aRepresentation)
	{	if (mState==RequestResponseFieldsImpl.EMPTY)
		{	mState = new RequestResponseFieldsImpl(10);
		}
		((RequestResponseFieldsImpl)mState).add(aKey, aRepresentation);
	}
	
	
	public void postCommissionSpace() throws Exception
	{	super.postCommissionSpace();
		bindIfNeeded();
		Thread.currentThread().setContextClassLoader(getClassLoader());
        for (int i=0; i<mMappings.length; i++)
        {	IStandardEndpoint mapping = mMappings[i];
			try
			{	mapping.postCommissionMapping(mStaticScope);
			}
			catch (Throwable e)
			{	ILogger logger=getKernel().getLogger();
				String msg=logger.format("MSG_STD_ENDPOINT_POSTCOMMISSION_FAILED",mapping.toString(),Utils.throwableToString(e));
				logger.logRaw(ILogger.WARNING,this,msg);
			}
        }
	}

	public void preDecommissionSpace() throws Exception
	{	Thread.currentThread().setContextClassLoader(getClassLoader());
		for (int i=0; i<mMappings.length; i++)
		{	IStandardEndpoint mapping = mMappings[i];
			try
			{	mapping.preDecommissionMapping(mStaticScope);
			}
			catch (Exception e)
			{	ILogger logger=getKernel().getLogger();
				String msg=logger.format("MSG_STD_ENDPOINT_PREDECOMMISSION_FAILED",mapping.toString(),Utils.throwableToString(e));
				logger.logRaw(ILogger.WARNING,this,msg);
			}
		}		
		super.preDecommissionSpace();
	}	

	public void onDecommissionSpace() throws Exception
	{	
		for (int i=0; i<mMappings.length; i++)
		{	IStandardEndpoint mapping = mMappings[i];
			try
			{	mapping.onDecommissionEndpoint();
			}
			catch (Exception e)
			{	ILogger logger=getKernel().getLogger();
				String msg=logger.format("MSG_STD_ENDPOINT_DECOMMISSION_FAILED",mapping.toString(),Utils.throwableToString(e));
				logger.logRaw(ILogger.WARNING,this,msg);
			}
		}
		super.onDecommissionSpace();
	}
	
	protected boolean isThisSpaceResolve(String aRequestId)
	{	return aRequestId.length()==0;
	}
	
	public void onResolve(INKFResolutionContext aContext) throws Exception
	{	
		if (mOwner.disableResolve()) return;
		
		if (aContext.isDebugResolve())
		{	onResolveDebug(aContext);
		}
		else
		{	INKFRequestReadOnly toResolve=aContext.getRequestToResolve();
			if (toResolve.getVerb()==INKFRequestReadOnly.VERB_META)
			{	String idString=toResolve.getIdentifier();
				if (isThisSpaceResolve(idString))
				{	// resolve for space meta
					aContext.createResolutionResponse(this);
				}
				else
				{	StandardSpaceMeta meta = innerGetSpaceMeta(aContext);
					Object[] mapping=meta.getMappingWithId(idString);
					if 	(mapping !=null )
					{	IStandardEndpoint endpoint=(IStandardEndpoint)mapping[0];
						String absolute=(String)mapping[1];
						IRequestScopeLevel parentScope = ((IRequest)aContext.getKernelContext().getThisKernelRequest().getValue(IRequest.TYPE_RESOLVE_REQUEST)).getRequestScope();
						aContext.createResolutionResponse(endpoint,absolute,getStaticScope(parentScope));
					}
					else
					{	aContext.createResponseFrom(null);
					}
				}
			}
			else
			{	//resolve for verbs other than META
				bindIfNeeded();
				for (int i=0; i<mMappings.length; i++)
				{	IStandardEndpoint mapping = mMappings[i];
					((NKFResolutionContextImpl)aContext).setEndpoint(mapping);
					mapping.onResolve(aContext);
					if (aContext.isResponseSet()) break;
				}
			}
		}
	}
	
	private void onResolveDebug(INKFResolutionContext aContext) throws Exception
	{
		INKFRequestReadOnly toResolve=aContext.getRequestToResolve();
		if (toResolve.getVerb()==INKFRequestReadOnly.VERB_META)
		{	String idString=toResolve.getIdentifier();
			if (isThisSpaceResolve(idString))
			{	// resolve for space meta
				aContext.createResolutionResponse(this);
				addResolveDebug(aContext,"STDR_CTX_META",null);
			}
			else
			{	StandardSpaceMeta meta = innerGetSpaceMeta(aContext);
			
				Object[] mapping=meta.getMappingWithId(idString);
				if 	(mapping !=null )
				{	IStandardEndpoint endpoint=(IStandardEndpoint)mapping[0];
					aContext.createResolutionResponse(endpoint);
					addResolveDebug(aContext,"STDR_PART_META",endpoint);
				}			
			}
		}
		else
		{	//resolve for verbs other than META		
			NKFResolutionContextImpl ctx=(NKFResolutionContextImpl)aContext;
			for (int i=0; i<mMappings.length; i++)
			{	IStandardEndpoint mapping = mMappings[i];
				ctx.setEndpoint(mapping);
				aContext.addResolutionDebug("**"); //this value is replaced
				mapping.onResolve(aContext);
				if (aContext.isResponseSet() )
				{	String msg=getKernel().getLogger().format("STDR_MATCH",mapping);
					aContext.insertResolutionDebug(msg, 0, null);
					break;
				}
				else
				{	String msg=getKernel().getLogger().format("STDR_NO_MATCH",mapping);
					aContext.insertResolutionDebug(msg, 0, null);
				}
			}
		}		
	}

	private void addResolveDebug(INKFResolutionContext aContext, String aMessageId, Object aOptionalParameter)
	{	String message;
		if (aOptionalParameter==null)
		{	message=getKernel().getLogger().format(aMessageId);
		}
		else
		{	message=getKernel().getLogger().format(aMessageId,aOptionalParameter.toString());
		}
		aContext.addResolutionDebug(message);
	}
	
	/** Append static scope on to end of a parent scope
	 * @param aParent
	 * @return
	 */
	private IRequestScopeLevel getStaticScope(IRequestScopeLevel aParent)
	{	IRequestScopeLevel scope = RequestScopeLevelImpl.createOrphanedRootScopeLevel(mStaticScopeArray[mStaticScopeArray.length-1], aParent);
		for (int i=mStaticScopeArray.length-2; i>=0; i--)
		{	scope = RequestScopeLevelImpl.createNonDurableScopeLevel(mStaticScopeArray[i], scope);
		}
		
		return scope;
	}
	
	private long mLast;
	private long mTotal;
	
	private StandardSpaceMeta innerGetSpaceMeta(INKFRequestContext aContext) throws Exception
	{	
		INKFResponseReadOnly resp=mLastMetaResponse;
		
		if (resp==null || resp.isExpired())
		{
			//shortcut evaluation of meta if this space is expiring too fast and too often
			final long PERIOD=200; //look for expiration over this period in ms
			final long COUNT=20; // look for this number of re-evaluations before taking action
			final long RETRY=100; // retry this number-COUNT times be re-evaluation
			
			long now=System.currentTimeMillis();
			if (now-mLast<PERIOD)
			{	mTotal++;
			}
			else
			{	mTotal=0;
			}
			mLast=now;
			
			if (mTotal<COUNT || mTotal==RETRY)
			{
				INKFRequest req=aContext.createRequest("");
				req.setVerb(INKFRequestReadOnly.VERB_META);
				req.setRepresentationClass(StandardSpaceMeta.class);
				resp = aContext.issueRequestForResponse(req);
				mLastMetaResponse=resp;
				/*
				IResponseMeta m=((NKFResponseReadOnlyImpl)resp).getKernelResponse().getMeta();
				if (m instanceof IInternalResponseMeta)
				{	IInternalResponseMeta im=(IInternalResponseMeta)m;
					int cost=im.getTotalCost();
					System.out.println("subtracting cost of space meta for "+this.toString()+" : "+cost);
					aContext.getKernelContext().declareDependency(ExpiryFactory.getNeverExpired(), Integer.MAX_VALUE, -cost);
				}
				*/
				if (mTotal==RETRY)
				{	mTotal=COUNT;
					//System.out.println("**RETRY META for "+toString());
				}
			}
			else
			{	if (mTotal==COUNT)
				{	//System.out.println("**SKIPPING META for "+toString());
				}
				
			}
		}
		StandardSpaceMeta result = (StandardSpaceMeta)resp.getRepresentation();
		return result;
	}
	
	public void onMeta(INKFRequestContext aContext) throws Exception
	{	INKFResponseReadOnly resp=mLastMetaResponse;
		if (resp!=null && !resp.isExpired())
		{	INKFResponse respOut=aContext.createResponseFrom(resp.getRepresentation());
			aContext.getKernelContext().declareDependency(resp);
		}
		else
		{	clearInitialisationProblems();
			TupleList partsTuples = new TupleList(3,32);
			IRequestScopeLevel scope = getStaticScope(aContext.getKernelContext().getThisKernelRequest().getRequestScope());
			StandardSpaceMeta result = new StandardSpaceMeta(mName,null,null,null);
			
			boolean bound=bindIfNeeded();
			for (int i=0; i<mMappings.length; i++ )
			{	IStandardEndpoint mapping = mMappings[i];
				try
				{	((NKFContextImpl)aContext).setRequestScope(scope);
					((NKFContextImpl)aContext).setEndpoint(mapping);
					ISpaceElements parts = mapping.getParts(aContext);
					if (parts!=null)
					{	if (parts.size()>0)
						{	for (int j=0; j<parts.size(); j++)
							{	IIdentifier partId = parts.getIdentifier(j);
								if (partId!=null)
								{	result.putMapping(partId.toString(), mapping);
								}
								IEndpoint ep=parts.getPhysicalEndpoint(j);
								if (ep==null) ep=mapping;
								partsTuples.put(partId,parts.getDelegatedSpace(j),ep);
							}
						}
						else
						{	partsTuples.put(null,null,mapping);
						}
					}
				}
				catch (Exception e)
				{	INetKernelException ex = Utils.createFormattedException("EX_STD_DURING_GETPARTS", "MSG_STD_DURING_GETPARTS", getKernel().getLogger(), e, mapping.toString(),this.toString());
					String msg=ex.toString();
					getKernel().getLogger().logRaw(ILogger.WARNING, this, msg);
					setInitialisationProblem(msg);
					aContext.getKernelContext().declareExpiryDependency();
				}
			}
			
			result.setParts(new SpaceElementsImpl(partsTuples));
			
			//state
			result.setAdditionalFields(mState);
			
			INKFResponse response =aContext.createResponseFrom(result);
			INKFExpiryFunction expiry=mInvalidationExpiry;
			if (expiry!=null)
			{	response.setExpiry(INKFResponse.EXPIRY_MIN_FUNCTION_DEPENDENT,expiry);
			}
		}
	}
	
	private void invalidateMeta(boolean aFinalInvalidation, boolean aNoInvalidation)
	{	
		if (mInvalidationExpiry!=null && !aNoInvalidation)
		{	mInvalidationExpiry.invalidate();
		}
		if (!aFinalInvalidation)
		{	mInvalidationExpiry = new GoldenThreadExpiryFunction("StandardSpace Binding Expiration");
		}
		else
		{	mInvalidationExpiry=null;
		}
	}
	
	/**
	 * @throws Exception
	 * @return true if bound
	 */
	private boolean bindIfNeeded() throws Exception
	{	boolean bound=mIsBound;
		if (!bound)
		{	boolean acquired=mBindingLock.compareAndSet(false, true);
			if (acquired)
			{	try
				{	int status=innerBind();
					if ((status&1)==1)
					{	bound=true;
						mIsBound=true;
					}
					if ((status&2)==2)
					{	invalidateMeta(bound,false);
					}
					else
					{	if (status==1)
						{	invalidateMeta(true,true);
						}
					}
				}
				finally
				{	mBindingLock.set(false);
				}
			}
		}
		return bound;
	}
	
	/**
	 * @return 0=still unbound, 1=nothing done complete, 2=work done but not bound, 3=bound complete, work done
	 * @throws Exception
	 */
	private int innerBind() throws Exception
	{
		NKFLifecycleEventContextImpl context = null;
		int bound=1;
		for (int i=0; i<mMappings.length; i++)
		{	IStandardEndpoint mapping = mMappings[i];
			if (mapping instanceof ILateBound)
			{	ILateBound lbp = (ILateBound)mapping;
				try
				{	if (!lbp.isBound())
					{	if (context==null)
						{	context = new NKFLifecycleEventContextImpl(getKernel(),mStaticScope,this,getParameters());
						}
						IStandardEndpoint boundMapping =lbp.bind(context);
						boundMapping.onCommissionEndpoint(getKernel(), this);
						mMappings[i]=boundMapping;
						mapping=mMappings[i];
						bound|=2;
					}
				}
				catch (INetKernelException e)
				{	Element config=null;
					if (mConfig!=null)
					{	config=XMLUtils.getFirstChildElement(mConfig);
						for (int j=0; j<i && config!=null; j++)
						{	config=XMLUtils.getNextSiblingElement(config);
						}
						if (config==null) config=mConfig;	
					}
					SpaceFactory.logConstructionError(e, config, (StandardSpace)getSpace(),getKernel());
					bound&=2;
				}
			}
		}
		return bound;
	}
	
	public String toString()
	{	return mName;
	}
}