package org.netkernel.module.standard.endpoint;

import org.netkernel.container.IKernel;
import org.netkernel.container.ILogger;
import org.netkernel.container.IMessages;
import org.netkernel.layer0.meta.IPrototypeParameterMeta;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.INKFResponseReadOnly;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.nkf.impl.INKFParametersControl;
import org.netkernel.layer0.nkf.impl.NKFContextImpl;
import org.netkernel.layer0.nkf.impl.NKFResponseImpl;
import org.netkernel.layer0.urii.SimpleIdentifierImpl;
import org.netkernel.layer0.util.OrderedCheapMultiMap;
import org.netkernel.layer0.util.TupleList;
import org.netkernel.layer0.util.XMLUtils;
import org.netkernel.module.standard.SpaceFactory;
import org.netkernel.module.standard.StandardSpace;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.ISpace;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.ParserConfigurationException;

public class Parameters implements INKFParametersControl
{
	private static final IPrototypeParameterMeta[] sNoParams = new IPrototypeParameterMeta[0];
	
	protected IPrototypeParameterMeta[] mParameterDefn;
	protected final OrderedCheapMultiMap mParameters;
	private boolean mTolerant;
	private IKernel mKernel;
	
	public static Parameters EMPTY = new Parameters();
	
	private Parameters()
	{	mParameterDefn = sNoParams;
		mParameters=OrderedCheapMultiMap.EMPTY;
		mTolerant=false;
	}
	/** Construct Parameters instance
	 * @param aParameterDefn an array of prototype parameters. An empty array specifies no parameters allowed,
	 *  null specifies anything goes
	 * @param aEndpoint
	 */
	public Parameters(IPrototypeParameterMeta[] aParameterDefn)
	{	mParameterDefn=aParameterDefn;
		int parameterCount=aParameterDefn.length;
		if (parameterCount==0)
		{	parameterCount=2;
		}
		mParameters=new OrderedCheapMultiMap(parameterCount);
	}
	
	private Parameters(Parameters aOther, IPrototypeParameterMeta[] aParameterDefn) //throws NKFException
	{	//mMessages=aOther.mMessages;
		int len = aOther.getParamCount();
		mParameters=new OrderedCheapMultiMap(len);
		for (int i=0; i<len; i++)
		{	mParameters.put(aOther.mParameters.getKey(i),aOther.mParameters.getValue(i));
		}
		mParameterDefn = StandardEndpointImpl.mergeParameterArray(aOther.mParameterDefn, aParameterDefn);
	}
	
	public INKFParametersControl cloneWithExtendedDefn(IPrototypeParameterMeta[] aParameterDefn) throws NKFException
	{	return new Parameters(this,aParameterDefn);
	}
	
	public void declareParameters(IPrototypeParameterMeta[] aParameterDefn)
	{
	}
	
	public void setTolerant(boolean aTolerant)
	{	mTolerant=aTolerant;
	}
	
	public boolean isTolerant()
	{	return mTolerant;
	}
		
	public void setParameterRaw(String aName, Object aValue)
	{	mParameters.put(aName, aValue);		
	}
	
	public void setParameter(String aName, Object aValue, IKernel aKernel) throws NKFException
	{	IPrototypeParameterMeta meta=null;
		try
		{	meta=getParamMeta(aName);
		}
		catch (Exception e)
		{	if (isTolerant())
			{	setParameterRaw(aName, aValue);
			}
			else
			{	throw generateParameterException(aName, aKernel.getLogger(), "MSG_STD_ENDPOINT_PARAM_UNDECLARED", null, null);
			}
		}
		if (meta!=null)
		{	Object value=convertParameterElement((Element)aValue, meta, aKernel);
			setParameterRaw(aName, value);
		}
	}
	
	private Object convertParameterElement(Element aValue, IPrototypeParameterMeta aMeta, IKernel aKernel) throws NKFException
	{	ILogger logger=aKernel.getLogger();
		Object result=null;
		try
		{	Class aRepresentation = aMeta.getRepresentationalClass();

			if (aRepresentation==ISpace.class || aRepresentation==Element.class)
			{	result=aValue;
			}
			else
			{	Element paramElement = aValue;
				String stringValue = XMLUtils.getText(paramElement);
				if (aRepresentation==String.class)
				{	result=stringValue;
				}
				else
				{	stringValue=stringValue.trim();
					if (aRepresentation==Integer.class)
					{	result=new Integer(stringValue);
					}
					else if (aRepresentation==Float.class)
					{	result=new Float(stringValue);
					}
					else if (aRepresentation==Long.class)
					{	result=new Long(stringValue);
					}
					else if (aRepresentation==Double.class)
					{	result=new Double(stringValue);
					}
					else if (aRepresentation==Boolean.class)
					{	boolean isTrue = stringValue.length()==0 || Boolean.valueOf(stringValue).booleanValue();
						result = Boolean.valueOf(isTrue);
					}
					else if (aRepresentation==IIdentifier.class)
					{	//if no children use text, if has children then use parameter element
						Element firstChild = XMLUtils.getFirstChildElement(paramElement);
						if(firstChild!=null)
						{	try
							{	Document d = XMLUtils.newDocument();
								d.appendChild(d.importNode(paramElement, true));
								result=d;
							}
							catch (ParserConfigurationException e)
							{	throw generateParameterException(aMeta.getName(),logger,"MSG_STD_ENDPOINT_PARAM_BAD_TYPE",null,null);
							}
						}
						else
						{	result=new SimpleIdentifierImpl(stringValue);
						}
					}
					else
					{	throw generateParameterException(aMeta.getName(),logger,"MSG_STD_ENDPOINT_PARAM_BAD_TYPE",null,null);
					}
				}
			}
			return result;
		}
		catch (java.lang.NumberFormatException e)
		{	e.printStackTrace();
			throw generateParameterException(aMeta.getName(),logger,"MSG_DECL_ENDPOINT_PARAM_PARSE",null,null);
		}
	}
	
	public void onCommission(IKernel aKernel, ISpace aSpace) throws NKFException
	{
		mKernel=aKernel;
		IMessages messages = aKernel.getLogger();
		
		//generate an sub-spaces now
		for (int i=0; i<mParameterDefn.length; i++)
		{	IPrototypeParameterMeta p = mParameterDefn[i];
			String name=p.getName();
			
			Class repClass = p.getRepresentationalClass();
			if (repClass==ISpace.class)
			{	try
				{	
					for(int j=0; j<mParameters.size(); j++)
					{	String eName=(String)mParameters.getKey(j);
						if (eName.equals(name))
						{	Object o = mParameters.getValue(j);
							if (o instanceof Element)
							{
								Element config = (Element)o;
								ISpace space = SpaceFactory.buildSubSpace(config,(StandardSpace)aSpace);
								mParameters.replaceValue(j,space);
							}
						}
					}
					//Element config=get
					//result = SpaceFactory.buildSubSpace(config,(StandardSpace)aSpace);
				}
				catch (Exception e)
				{	throw generateParameterException(p.getName(),messages,"MSG_STD_ENDPOINT_PARAM_SPACE",null,e);
				}
			}
			
			int size=mParameters.size(name);
			//verify that we have as many as we need
			if (size<p.getMinimumValency())
			{	throw generateParameterException(name,messages,"MSG_STD_ENDPOINT_PARAM_VALENCY_MIN");
			}
			// verify we don't have too many
			if (size>p.getMaximumValency())
			{	throw generateParameterException(name,messages,"MSG_STD_ENDPOINT_PARAM_VALENCY_MAX");
			}
			// verify that we have default if we don't maximum
			if (p.getMaximumValency()>size && p.getDefault()==IPrototypeParameterMeta.NO_DEFAULT)
			{	throw generateParameterException(name,messages,"MSG_STD_ENDPOINT_PARAM_NO_DEFAULT");
			}
			if (size==0 && p.getDefault()==IPrototypeParameterMeta.NO_DEFAULT)
			{	throw generateParameterException(name,messages,"MSG_STD_ENDPOINT_PARAM_NO_DEFAULT");
			}
		}
		
		if (!mTolerant)
		{	for (int i=0; i<mParameters.size(); i++)
			{	String name = (String)mParameters.getKey(i);
				IPrototypeParameterMeta parameter=getParamMeta(name);
				if (parameter==null)
				{	throw generateParameterException(name,messages,"MSG_STD_ENDPOINT_PARAM_UNDECLARED");
				}		
			}
		}
	}
	
	private NKFException generateParameterException(String aName, IMessages aMessages, String aMessageId )
	{	return generateParameterException(aName,aMessages,aMessageId,null,null);
	}
	
	private NKFException generateParameterException(String aName, IMessages aMessages, String aMessageId, Object aArg, Exception e)
	{	String id = aMessages.format("EX_STD_ENDPOINT_PARAM");
		String message;
		if (aArg!=null)
		{ message = aMessages.format(aMessageId, aName, aArg);
		}
		else
		{ message = aMessages.format(aMessageId, aName);
		}
		return new NKFException(id,message,e);
	}
	
	public IPrototypeParameterMeta getParamMeta(String aName) throws NKFException
	{	IPrototypeParameterMeta result=innerGetParameterMeta(aName);
		
		if (result==null)
		{	throw new NKFException("Use of undeclared parameter",aName);
		}
		return result;
	}
	
	private IPrototypeParameterMeta innerGetParameterMeta(String aName)
	{	IPrototypeParameterMeta result=null;
		for (int i=0; i<mParameterDefn.length; i++)
		{	IPrototypeParameterMeta p = mParameterDefn[i];
			if (p.getName().equals(aName))
			{	result=p;
			}
		}
		return result;
	}

	public int getParamCount()
	{	return mParameters.size();
	}
	
	public String getParamName(int aIndex)
	{	return (String)mParameters.getKey(aIndex);
	}
	
	public Object getParamValue(String aName) throws NKFException
	{	Object value=mParameters.get(aName);
		value=lateConvertValue(value,aName);
		return value;
	}
	
	public Object getParamValue(int aIndex) throws NKFException
	{	Object value=mParameters.get(aIndex);
		String name=(String)mParameters.getKey(aIndex);
		value=lateConvertValue(value,name);
		return value;		
	}
	
	public boolean paramExists(String aName)
	{	boolean exists;
		IPrototypeParameterMeta meta = innerGetParameterMeta(aName);
		if (meta!=null)
		{	Object d=meta.getDefault();
			exists = (d!=null && d!=IPrototypeParameterMeta.NO_DEFAULT ) || mParameters.exists(aName) ;
		}
		else
		{	exists = false;
		}
		return exists;
	}
	
	public List getParamValues(String aName) throws NKFException
	{	int len=mParameters.size();
		List result = new ArrayList(mParameters.size());		
		for (int i=0; i<len; i++)
		{	if (mParameters.getKey(i).equals(aName))
			{	Object value=mParameters.getValue(i);
				value=lateConvertValue(value,aName);
				result.add(value);
			}
		}
		return result;
	}
	
	private Object lateConvertValue(Object aValue, String aName) throws NKFException
	{	Object result;
		if (aValue==null)
		{	IPrototypeParameterMeta meta=getParamMeta(aName);
			result=meta.getDefault();
		}
		else
		{	result=aValue;
			if (aValue instanceof Element)
			{	IPrototypeParameterMeta meta=getParamMeta(aName);
				if (meta.getRepresentationalClass()!=Element.class)
				{	result=convertParameterElement((Element)aValue, meta, mKernel);
				}
			}
		}
		return result;
	}
	
	public INKFResponseReadOnly getParameterLiteralAs(String aName, Class aDesiredRepresentation, INKFRequestContext aContext) throws NKFException
	{	INKFResponseReadOnly result=null;
		Object representation = getParamValue(aName);
		if (mTransreptedParameterLiteralCache!=null)
		{	for (int i=0; i<mTransreptedParameterLiteralCache.size(); i++)
			{	if (mTransreptedParameterLiteralCache.getValue(i, 0)==aDesiredRepresentation && mTransreptedParameterLiteralCache.getValue(i, 1).equals(aName))
				{	result = (INKFResponseReadOnly)mTransreptedParameterLiteralCache.getValue(i, 2);
					if (result.isExpired()) result=null;
					break;
				}
			}
		}
		else
		{	mTransreptedParameterLiteralCache = new TupleList(3,4);
		}
		
		if (result==null)
		{	INKFRequest req=aContext.createRequest("param-literal:"+aName);
			req.setVerb(INKFRequestReadOnly.VERB_TRANSREPRESENT);
			NKFResponseImpl resp = new NKFResponseImpl(representation,(NKFContextImpl)aContext);
			resp.setExpiry(INKFResponse.EXPIRY_NEVER);
			req.addPrimaryArgumentFromResponse(resp);
			req.setRepresentationClass(aDesiredRepresentation);
			result = aContext.issueRequestForResponse(req);
			mTransreptedParameterLiteralCache.put(aDesiredRepresentation,aName,result);
		}
		else
		{	aContext.getKernelContext().declareDependency(result);
		}
		return result;
	}
	
	private TupleList mTransreptedParameterLiteralCache=null;
	
}
